// plates.inc - hull plating for spherical/cylindrical starship parts.
//              Also works well for brick columns, domed ceilings, etc.
//
// Author:   Jeff Lee <shipbrk@gate.net>
// Version:  1.2
// Date:     Friday, 2 July 1999
// Platform: POV-Ray 3.1
//
// Credits:  Thanks to Ron Parker <parkerr@fwi.com> for suggesting the method
//           used to overcome the blend_map limitation.
//
// If this file was distributed without documentation, you may obtain the
// documentation, a sample scene and the latest version of this file from:
//
//       http://www.gate.net/~shipbrk/raytrace/plates.html
//
// Version 1.2:  2 Jul 1999 -- Blend_map trickery caused immense slowdown in
//                             rendering!  Added alternate code to speed up
//                             the textures when everything fits within a
//                             single blend_map.  Textures which exceed the
//                             limit are still slow.
//
// Version 1.1:  1 Jul 1999 -- Fixed blend_map limitation, removed texture{}
//                             wrapper so that turbulence and warps can be
//                             applied.  Fixed spherical texture positioning
//                             bug.  (Unreleased version)
//
// Version 1.0: 30 Jun 1999 -- Initial release.

#ifndef (safeatan) #macro safeatan(a,b) #if (a|b) degrees(atan2(a,b)) #else 0.0 #end #end #end

//============================================================================
// MakeSphericalPlates() macro
//
// MainRad:  Radius of target object
// PWidth:   Width of plates (at widest point)
// PHeight:  Height of plates
// JWidth:   Width (and height) of the joining seams ("mortar").
// PlateTex: Base texture of plates.  Should be aligned in XY plane, pointing
//           towards -Z; will be rotated to lie on the face of each individual
//           plate.
// JoinTex:  Base texture of the plate joins (will not be rotated).
//
#macro MakeSphericalPlates(MainRad,PWidth,PHeight,JWidth,PlateTex,JoinTex)
  #ifndef (platedebug) #local platedebug=0; #end
  #local TexOfs = 0;
  #local UpperLimit=63;
  #if (JWidth = 0) #local UpperLimit=126; #end
  #local ShiftPlates = 0;
  #local foo = MainRad * 0.5 * pi;
  #local NumPlates = int(foo / (PHeight+JWidth));
  #local Lat = 1;
  #while (NumPlates/Lat > UpperLimit) #declare Lat = Lat+1;  #end
  #local RowTemp  = array[Lat]
  #local RowStart = array[Lat]
  #local RowEnd   = array[Lat]
  #local PlateHeight = (foo - (NumPlates*JWidth))  / NumPlates;
  #local HalfJoinHeight = JWidth / foo / 2;
  #local VertIter = 0;
  #local OuterLoop = 0;
  #local LatIndex = 0;
  #while (VertIter < NumPlates)
    #local RowStart[LatIndex] = cos( radians( 90-(90*VertIter/NumPlates) ));
    #local RowTemp[LatIndex] = texture {
      gradient y
      texture_map {
        #local RowOK = 1;
        #while (RowOK)
          #local BottomPoint = cos( radians( 90-(90*VertIter/NumPlates)-(90*HalfJoinHeight) ));
          #local TopPoint    = cos( radians( 90-(90*(VertIter+1)/NumPlates)+(90*HalfJoinHeight) ));
          #local MedPointX   = sin( radians( 90-(90*(VertIter+0.5)/NumPlates)));
          #local MedPointY   = cos( radians( 90-(90*(VertIter+0.5)/NumPlates)));
          #local PlateRotX   = 90-(safeatan( MedPointX, MedPointY ));
          #local PlateTranZ  = -(MainRad * MedPointY);
          #if (VertIter = NumPlates-1) #local PlateRotX=90; #end
          #local LevelCirc  = MainRad * pi * 2 * sin( radians( 90-(90*VertIter/NumPlates)-(90*HalfJoinHeight)));
          #local RowPlates  = int(LevelCirc / (PWidth+JWidth));
          #if (RowPlates < 1) #local RowPlates=1; #end
          #if (platedebug)
            #debug concat( "MakeSphericalPlates: row ", str(VertIter+1,0,0), " of ", str(NumPlates,0,0), " - ", str(RowPlates,0,0), " plates\n" )
          #end
          #local Freq = 1;
          #while (RowPlates/Freq > UpperLimit) #declare Freq = Freq + 1; #end
          #local CylTemp  = array[Freq]
          #local CylStart = array[Freq]
          #local CylEnd   = array[Freq]
          #local PlateWidth = (LevelCirc - (RowPlates*JWidth))  / RowPlates;
          #local HalfJoinWidth = JWidth / LevelCirc / 2;
          #local HorIter = 0;
          #local InnerLoop = 0;
          #local InTexIndex = 0;
          #while (HorIter < RowPlates)
            #if (InnerLoop = 0)
              #local CylStart[InTexIndex] = (HorIter/RowPlates)+HalfJoinWidth;
              #local CylTemp[InTexIndex] = texture {
                radial
                texture_map {
                  #local CylOK = 1;
                  #while (CylOK)
                    #local RightPoint = (HorIter/RowPlates)+HalfJoinWidth;
                    #local LeftPoint  = ((HorIter+1)/RowPlates)-HalfJoinWidth;
                    #local MedPointX  = sin( radians( (360*(HorIter+0.5)/RowPlates)));
                    #local MedPointY  = cos( radians( (360*(HorIter+0.5)/RowPlates)));
                    #local PlateRotY  = (safeatan( MedPointX, MedPointY ));
                    #if (PlateRotY<0) #local PlateRotY=360+PlateRotY; #end
                    #local NewXTex = texture { PlateTex translate <TexOfs,-TexOfs,PlateTranZ> scale 1/MainRad rotate <-PlateRotX,PlateRotY-90,0> }
                    #if (JWidth != 0) [RightPoint JoinTex] #end
                    [RightPoint NewXTex]
                    [LeftPoint  NewXTex]
                    #if (JWidth != 0) [LeftPoint  JoinTex] #end
                    #local InnerLoop = InnerLoop+1;
                    #local HorIter=HorIter+1;
                    #local TexOfs=TexOfs+1234;
                    #if (InnerLoop >= UpperLimit) #local CylOK = 0;  #end
                    #if (HorIter >= RowPlates) #local CylOK = 0;  #end
                  #end
                }
              }
              #local CylEnd[InTexIndex] = LeftPoint;
              #local InTexIndex = InTexIndex + 1;
              #local InnerLoop = 0;
            #end
          #end
          #local NewYTex = texture { 
            #if (Freq = 1)
              CylTemp[0]
            #else
              radial
              texture_map {
                #local InnerLoop = 0;
                #while (InnerLoop < Freq)
                    [CylStart[InnerLoop] CylTemp[InnerLoop]]
                    [CylEnd[InnerLoop]   CylTemp[InnerLoop]]
                    #local InnerLoop = InnerLoop + 1;
                #end
              }
            #end
            #if (VertIter=0)
              #local BottomRot = 360*((0.5/RowPlates)); // +HalfJoinWidth);
              #if (mod(RowPlates,2)=1) #local BottomRot=BottomRot*2; #end
            #end
            rotate ShiftPlates*y*45
            #local ShiftPlates=1-ShiftPlates;
          }
          #if (JWidth != 0) [BottomPoint JoinTex] #end
          [BottomPoint NewYTex]
          [TopPoint    NewYTex]
          #if (JWidth != 0) [TopPoint    JoinTex] #end
          #local VertIter=VertIter+1;
          #local OuterLoop = OuterLoop + 1;
          #if (OuterLoop >= UpperLimit) #local RowOK = 0;  #end
          #if (VertIter >= NumPlates) #local RowOK = 0; #end
        #end
      }
    }
    #local RowEnd[LatIndex] = cos( radians( 90-(90*(VertIter)/NumPlates) ));
    #local LatIndex = LatIndex + 1;
    #local OuterLoop = 0;
  #end
  #local T_Sphere1 = texture {
    #if (Lat = 1)
      RowTemp[0]
    #else
      gradient y
      texture_map {
        #local OuterLoop = 0;
        #while (OuterLoop < Lat)
          [RowStart[OuterLoop] RowTemp[OuterLoop]]
          [RowEnd[OuterLoop]   RowTemp[OuterLoop]]
          #local OuterLoop = OuterLoop + 1;
        #end
      }
    #end
    rotate y*-90 // Put the seams in +Z
    scale <1,0.5,1>
  }
  gradient y
  texture_map {
    [0.5 T_Sphere1 translate y*1 rotate y*BottomRot]
    [0.5 T_Sphere1 rotate x*180 ]
  }
  translate y*-0.5
  scale <1,2,1>*MainRad
#end

//============================================================================
// MakeCylindricalPlates() macro
//
// Same parameters as MakeSphericalPlates().
//
#macro MakeCylindricalPlates(MainRad,PWidth,PHeight,JWidth,PlateTex,JoinTex)
  #local T_Cylinder1 = texture {
    #local UpperLimit=63;
    #if (JWidth = 0) #local UpperLimit=126; #end
    #local TexOfs = 0;
    #local LevelCirc  = MainRad * pi * 2;
    #local RowPlates  = int(LevelCirc / (PWidth+JWidth));
    #local Freq = 1;
    #while (RowPlates/Freq > UpperLimit) #declare Freq = Freq + 1; #end
    #local CylTemp  = array[Freq]
    #local CylStart = array[Freq]
    #local CylEnd   = array[Freq]
    #local PlateWidth = (LevelCirc - (RowPlates*JWidth))  / RowPlates;
    #local HalfJoinWidth = JWidth / LevelCirc / 2;
    #local HorIter = 0;
    #local OffsetRot = 360/RowPlates/2;
    #if (mod(RowPlates,2)=1) #local OffsetRot=OffsetRot*2; #end
    #local InnerLoop = 0;
    #local TexIndex = 0;
    #while (HorIter < RowPlates)
      #local CylStart[TexIndex] = (HorIter/RowPlates)+HalfJoinWidth;
      #local CylTemp[TexIndex] = texture {
        radial
        texture_map {
          #local CylOK = 1;
          #while (CylOK)
            #local RightPoint = (HorIter/RowPlates)+HalfJoinWidth;
            #local LeftPoint  = ((HorIter+1)/RowPlates)-HalfJoinWidth;
            #local MedPointX  = sin( radians( (360*(HorIter+0.5)/RowPlates)));
            #local MedPointY  = cos( radians( (360*(HorIter+0.5)/RowPlates)));
            #local PlateRotY  = (safeatan( MedPointX, MedPointY ));
            #if (PlateRotY<0) #local PlateRotY=360+PlateRotY; #end
            #local NewXTex = texture { PlateTex scale <1/MainRad,1/PHeight,1/MainRad> translate <TexOfs,-TexOfs,0> rotate y*(PlateRotY-90) }
            #if (JWidth != 0) [RightPoint JoinTex] #end
            [RightPoint NewXTex]
            [LeftPoint  NewXTex]
            #if (JWidth != 0) [LeftPoint  JoinTex] #end
            #local InnerLoop = InnerLoop+1;
            #local HorIter=HorIter+1;
            #local TexOfs=TexOfs+1234;
            #if (InnerLoop >= UpperLimit) #local CylOK = 0;  #end
            #if (HorIter >= RowPlates) #local CylOK = 0;  #end
          #end
        }
      }
      #local CylEnd[TexIndex] = LeftPoint;
      #local TexIndex = TexIndex + 1;
      #local InnerLoop = 0;
    #end
    #if (Freq = 1)
      CylTemp[0]
    #else
      radial
      texture_map {
        #local InnerLoop = 0;
        #while (InnerLoop < Freq)
          [CylStart[InnerLoop] CylTemp[InnerLoop]]
          [CylEnd[InnerLoop]   CylTemp[InnerLoop]]
          #local InnerLoop = InnerLoop + 1;
        #end
      }
    #end
  }
  #local BottomPoint = (JWidth / PHeight) / 2;
  #local TopPoint = 1-BottomPoint;
  #local T_Cylinder2 = texture {
    gradient y
    texture_map {
      #if (JWidth != 0) [BottomPoint JoinTex] #end
      [BottomPoint T_Cylinder1]
      [TopPoint    T_Cylinder1]
      #if (JWidth != 0) [TopPoint    JoinTex] #end
    }
    scale <1,0.5,1>
  }
  gradient y
  texture_map {
    [0.5 T_Cylinder2 ]
    [0.5 T_Cylinder2 translate y*30 rotate y*(180+OffsetRot)]
  }
  warp { repeat y offset y*50 }
  scale <MainRad,2*PHeight,MainRad>
#end
